/*
 *  Shader.cpp
 *  metricspace
 *
 *  Created by Michael Tauraso on 12/19/06.
 *  Copyright 2006 __MyCompanyName__. All rights reserved.
 *
 */



#include <stdexcept>
#include <algorithm>

#include "Shader.h"

#include "FileParser.h"

std::vector<ShaderProg*> * ShaderProg::s_shaderProgs = NULL;
ShaderProg * ShaderProg::s_shaderInUse = NULL;

void Shader::dump(std::ostream& out)
{
	GLint logLength = 0;
	glGetObjectParameterivARB(m_shader,GL_OBJECT_INFO_LOG_LENGTH_ARB,&logLength);
	
	if(logLength == 0)
	{
		out << "No info log for shader " << m_filename <<"\n";
		return;
	}
	char * infoLog = new char[logLength];
	
	glGetInfoLogARB(m_shader, logLength, &logLength, infoLog);
	
	out << "Dumping info log from shader " << m_filename 
		<< " (length " << logLength<<") :\n";
	out.write(infoLog, logLength);
	
	delete[] infoLog;
}

Shader::Shader(std::string filename, GLenum shaderType)
{
	Init(filename, shaderType);
}

Shader::Shader(const Shader& shader)
{
	//get our type
	GLint type;
	
	glGetObjectParameterivARB(m_shader,GL_OBJECT_SUBTYPE_ARB, &type);
	
	Init(shader.m_filename,type);
}

void Shader::Init(std::string filename, GLenum shaderType)
{
	// actually make the shader
	m_shader = glCreateShaderObjectARB(shaderType);
	
	//load up our main source
	size_t size;
	const char ** source = FileParser::parseLines(filename,size);
	glShaderSourceARB(m_shader, size, source, NULL);
	FileParser::freeLines(source, size);
	
	//keep our source around
	m_filename = filename;
	
	//compile us up
	compile();
}

Shader::~Shader()
{
	glDeleteObjectARB(m_shader);
}

std::string Shader::getFilename()
{
	return m_filename;
}

void Shader::compile()
{
	GLint result;
	
	glCompileShaderARB(m_shader);
	glGetObjectParameterivARB(m_shader, GL_OBJECT_COMPILE_STATUS_ARB, &result);
	if(!result)
	{
		//bail here, we have a compile error
		dump(std::cerr);
		throw std::runtime_error("Shader::compile() error: " + m_filename);
	}
}


//ShaderProg


ShaderProg::ShaderProg()
{
	//keep track of us
	if(s_shaderProgs == NULL)
	{
		s_shaderProgs = new std::vector<ShaderProg*>();
	}
	s_shaderProgs->push_back(this);
	
	//create us
	m_shader = glCreateProgramObjectARB();
}

ShaderProg::~ShaderProg()
{
	std::vector<ShaderProg*>::iterator me;
	me = find(s_shaderProgs->begin(),s_shaderProgs->end(), this);
	s_shaderProgs->erase(me);
	
	if(s_shaderProgs->size() == 0)
	{
		delete s_shaderProgs;
		s_shaderProgs = NULL;
	}
}

GLhandleARB ShaderProg::getHandle()
{
	return m_shader;
}


void ShaderProg::dump(std::ostream & out)
{
	//first dump us
	out << "Dump of ShaderProg object:\n";
	
	GLint logLength;
	glGetObjectParameterivARB(m_shader,GL_OBJECT_INFO_LOG_LENGTH_ARB,&logLength);
	
	if(logLength != 0)
	{

		char * infoLog = new char[logLength];
		
		glGetInfoLogARB(m_shader, logLength, &logLength, infoLog);
		
		out << "Dumping info log from ShaderProg:\n";
		out.write(infoLog, logLength);
		
		delete[] infoLog;
	}
	else
	{
		out << "No error log to dump for this ShaderProg\n";
	}
	
	
	//dump things attached to us
	for(std::vector<Shader* >::iterator i = m_shaders.begin(); 
		i != m_shaders.end(); i++)
	{
		(*i)->dump(out);
	}
	
}

void ShaderProg::attach(Shader * shader)
{
	m_shaders.push_back(shader);
	glAttachObjectARB(m_shader, shader->m_shader);
}

void ShaderProg::detach(std::string filename)
{
	std::vector<Shader * >::iterator i;
	for(i = m_shaders.begin(); i!= m_shaders.end(); i++)
	{
		if((*i)->getFilename() == filename)
		{
			break;
		}
	}
	
	if(i != m_shaders.end())
	{
		glDetachObjectARB(m_shader, (*i)->m_shader); 
		delete *i;
		m_shaders.erase(i);
	}
}

void ShaderProg::link()
{
	GLint result;
	
	glLinkProgramARB(m_shader);
	glGetObjectParameterivARB(m_shader, GL_OBJECT_LINK_STATUS_ARB, &result);
	if(!result)
	{
		//bail here, we have a link error
		dump(std::cerr);
		throw std::runtime_error("ShaderProg::link() error");
	}
}

/*void ShaderProg::validate()
{
	GLint result;
	glValidateProgramARB(m_shader);
	glGetObjectParameterivARB(m_shader, GL_OBJECT_VALIDATE_STATUS_ARB, &result);
	if(!result)
	{
		// bail for some reason we can't validate... what does this mean
		// anyway
		dump(std::cerr);
		throw std::runtime_error("ShaderProg::validate() error");
	}
}*/

void ShaderProg::use()
{
	//GLint result;
	
	//link();
	//validate();
	
	std::cerr << "Using shader dumped below:\n";
	dump(std::cerr);
	
	s_shaderInUse = this;
	glUseProgramObjectARB(m_shader);
	
	GLhandleARB inUse = glGetHandleARB(GL_PROGRAM_OBJECT_ARB);
	
	if(inUse != m_shader)
	{
		std::cerr<< "Aww fuck and we aren't even using the shader\n";
		std::cerr<< "maybe you should have linked it\n";
	}
	
}




